@@ -1,29 +1,47 @@
-import heapq
-import threading
+__all__ = ["SNMPv1Manager"]
 
-from snmp.dispatcher import *
+import heapq
+import math
+import threading
+import time
+
+from snmp.manager import *
 from snmp.message import *
 from snmp.pdu import *
-from . import *
+from snmp.utils import *
+
+class Request(RequestHandle):
+    def __init__(self, pdu, manager, community,
+                timeout=10.0, refreshPeriod=1.0):
+        now = time.time()
 
-class Request(Dispatcher.Handle):
-    def __init__(self, pdu, manager, community):
         self.community = community
         self.manager = manager
         self.pdu = pdu
 
         self.callback = None
         self.event = threading.Event()
-        self.expired = False
         self.response = None
 
+        self.expiration = now + timeout
+        self._nextRefresh = self.expiration
+        self.period = refreshPeriod
+
     def __del__(self):
         if self.callback is not None:
             self.close()
 
     @property
-    def fulfilled(self):
-        return self.event.is_set()
+    def expired(self):
+        return self.expiration <= time.time()
+
+    @property
+    def nextRefresh(self):
+        return self._nextRefresh
+
+    @nextRefresh.setter
+    def nextRefresh(self, value):
+        self._nextRefresh = min(self.expiration, value)
 
     def close(self):
         self.callback(self.pdu.requestID)
@@ -39,8 +57,30 @@ class Request(Dispatcher.Handle):
         self.response = response
         self.event.set()
 
+    def reallySend(self):
+        self.manager.sendPdu(self.pdu, self, self.community)
+
+    def refresh(self):
+        if self.event.is_set():
+            return None
+
+        now = time.time()
+        delta = self.nextRefresh - now
+
+        if delta <= 0.0:
+            if self.expiration <= now:
+                return None
+
+            self.nextRefresh += math.ceil(-delta / self.period) * self.period
+            self.reallySend()
+            return 0.0
+        else:
+            return delta
+
     def send(self):
-        return self.manager.sendPdu(self.pdu, self, self.community)
+        now = time.time()
+        self.nextRefresh = now + self.period
+        self.reallySend()
 
     def wait(self):
         pdu = None
@@ -57,11 +97,11 @@ class Request(Dispatcher.Handle):
             return pdu
 
 class SNMPv1Manager:
-    def __init__(self, engine, locator, community, autowait=True):
+    def __init__(self, dispatcher, locator, community, autowait=True):
         self.autowait = autowait
         self.locator = locator
 
-        self.localEngine = engine
+        self.dispatcher = dispatcher
         self.defaultCommunity = community
 
         self.lock = threading.Lock()
@@ -70,27 +110,31 @@ class SNMPv1Manager:
     def refresh(self):
         while self.requests:
             with self.lock:
-                entry = self.requests[0]
-                result = entry.refresh()
+                reference = self.requests[0]
+                request = reference()
 
-                if result is None:
+                if request is None:
+                    wait = None
+                else:
+                    wait = request.refresh()
+
+                if wait is None:
                     heapq.heappop(self.requests)
                     continue
-
-                if result < 0:
-                    heapq.heapreplace(self.requests, entry)
+                elif wait > 0.0:
+                    return wait
                 else:
-                    return result
+                    heapq.heapreplace(self.requests, reference)
 
         return None
 
     def sendPdu(self, pdu, handle, community):
-        return self.localEngine.sendPdu(
+        self.dispatcher.sendPdu(
             self.locator,
             MessageProcessingModel.SNMPv1,
             pdu,
             handle,
-            community
+            community,
         )
 
     def sendRequest(self, pdu, community=None, wait=None, **kwargs):
@@ -101,9 +145,10 @@ class SNMPv1Manager:
             wait = self.autowait
 
         request = Request(pdu, self, community, **kwargs)
+        reference = ComparableWeakRef(request, key=lambda r: r.nextRefresh)
 
         with self.lock:
-            heapq.heappush(self.requests, Repeater(request))
+            heapq.heappush(self.requests, reference)
             request.send()
 
         if wait:
